# -*- coding: utf-8 -*-
"""Module to extract tables from text

get_table(string, int, boolean, int)

Table class methods:
   __init__(string, int, string)
   get_cols()
   get_headers_text(int) 
   get_rows()
   get_row(Loc) -> Loc list
   get_colnumbers(row) -> int list
   untag_rows()
"""
import regex as re

from .extracterror import handle_error
from .loc import expand, Loc
from .loctuple import intersect_len
from .loctuple import first, subinterval, before, intersects, seq_before, meets
from .loclist import binary_search
from .tagger import Tagger
from .loctuplelist import groupby
from . import utilities as ut
class Table: 
   """Table object
   Constructor Parameters:
      text (string) -- text where table is
      hlinecnt (int) -- number of lines in the table header
      fieldRE (string) -- regular expression that matches fields in the table (default ut.FIELD)
   Set data member tagged to define tags on text
   """
   def __init__(self, text, hlinecnt, fieldRE = ut.FIELD):
      non_empty = []
      for m in re.finditer(ut.LINE, text):
         if re.search('\S', m.group()) != None:
            # line with non-space characters           
            if hlinecnt == 0 or len(non_empty) == hlinecnt-1:
               non_empty.append((m.span()[0],len(text)))
               break
            else:
               non_empty.append(m.span())
      if hlinecnt != 0 and len(non_empty) != hlinecnt:
         fmt = "Table does not have enough header lines: {}"
         handle_error(110701, fmt.format(text))
      table_text = ''
      for fr,to in non_empty:
         table_text = table_text + text[fr:to]
      self.hlinecnt = hlinecnt
      self.tagged = Tagger(table_text)
      self.tagged.tagRE('LINE', ut.LINE)                  # line
      self.tagged.tagRE('FIELD', fieldRE, 2)              # field
      # get headers -- first <hlinecnt> lines
      locs = self.tagged.get_locs('LINE')
      if hlinecnt > 0:
         self.tagged.tag_list('lHDER', locs[:hlinecnt] )  # header line
      self.tagged.tag_list('lCONT', locs[hlinecnt:])      # content line
      # project fields to beginning of line
      # (i.e. relative to beginning of line)
      locs = self.tagged.project('FIELD', 'LINE')
      self.tagged.tag_list('pFIELD', locs)                # projected field
      tuples = self.tagged.select(subinterval, ['FIELD', 'lHDER'])
      hdr_field = list(map(first, tuples))
      tuples = self.tagged.select(subinterval, ['FIELD', 'lCONT'])
      cont_field = list(map(first, tuples))
      self.tagged.tag_list('cFIELD', cont_field)          # content field
      self.tagged.tag_list('hFIELD', hdr_field)           # header field
      locs = self.tagged.project('cFIELD', 'LINE')
      self.tagged.tag_list('pcFIELD', locs )              # projected content field
      locs = self.tagged.project('hFIELD', 'LINE')
      self.tagged.tag_list('phFIELD', locs)               # projected header field
   def get_cols(self):
      """Get columns from table

      Set the following data members:
         cols_count -- number of columns in table 
         cols_lengths -- list of column lengths
         cont_cols_count -- number of content columns
         cont_cols_lengths -- list of content column lengths
         hdr_cols_count -- number of header columns
         hdr_cols_lengths -- list of header column lengths 
      Tag columns as 'COL_0', 'COL_1', ....
      """
      self.cols = ut.cluster(self.tagged.get_locs('pFIELD'))
      self.abs_cols = [sorted(list(map(expand, column)), key=lambda x:x.order()) \
                       for column in self.cols] 
      # tags columns as COL_0, COL_1, ...
      self.tagged.tag_lists('COL', self.abs_cols)
      self.cols_count = len(self.cols)
      self.cols_lengths = [len(col) for col in self.cols]

      self.cont_cols = ut.cluster(self.tagged.get_locs('pcFIELD'))
      self.cont_cols_count = len(self.cont_cols)
      self.cont_cols_lengths = [len(col) for col in self.cont_cols]
 
      self.hdr_cols = ut.cluster(self.tagged.get_locs('phFIELD'))
      self.hdr_cols_count = len(self.hdr_cols)
      self.hdr_cols_lengths = [len(col) for col in self.hdr_cols] 
   
   def split_cont_cols(self):
      """
      split a content column that align with 2 or more headers
      """
      def get_col_to_split():
         cont_spans = [ut.get_list_span(col) for col in self.cont_cols]
         hdr_spans = [ut.get_list_span(col) for col in self.hdr_cols]
         score = []
         hdr_cols = {}
         for cc in range(self.cont_cols_count):
            cont_span = cont_spans[cc]
            if cont_span is None:
               score.append(-1)
               continue
            hdr_cols[cc] = []
            for hc in range(self.hdr_cols_count):
               if intersects( (cont_span, hdr_spans[hc]) ):
                  hdr_cols[cc].append(hc)
            if len(hdr_cols[cc]) < 2: 
               score.append(-1)
               continue
            len_cont_span = cont_span.length()
            int_perc = sum([intersect_len((cont_span, hdr_spans[j])) \
                            for j in hdr_cols[cc]]) / len_cont_span 
            score.append(1-int_perc)
         max_score = max(score)
         if max_score < 0: return None, None
         cc_best = [k for k in range(self.cont_cols_count) if score[k] == max_score][0]
         hdr_spans = [ut.get_list_span(self.hdr_cols[col]) for col in hdr_cols[cc_best]]
         return cc_best, hdr_spans
      def partition(cont_loc, hdr_spans, j):
         if j == len(hdr_spans)-1 or not intersects((cont_loc, hdr_spans[j+1])):
            return cont_loc, []
         cont_txt = self.tagged.get_text_loc(cont_loc)
         space_at = [i for i,ch in enumerate(cont_txt) if ch == ' ']
         for i in space_at:
            curr_loc = Loc(cont_loc.start(), cont_loc.start()+i, cont_loc.offset)
            next_loc = Loc(cont_loc.start()+i+1, cont_loc.end(), cont_loc.offset)
            if not intersects((next_loc, hdr_spans[j])) and intersects((next_loc, hdr_spans[j+1])):
               return curr_loc, [next_loc]
         return cont_loc, []
      def split_cont_col(col2split,hdr_spans ):
         """
            split content column nr col2split into n columns where n is the
            number of elements in hdr_spans
            hdr_spans are the spans of headears that align with column col2split
            
         """
         cols = [ [] for i in range(len(hdr_spans)) ] # cols to replace self.cont_cols[cc_best]
         if len(self.tagged.get_locs('TOSPLIT')) != 0:
            self.tagged.del_tag('TOSPLIT')
         self.tagged.tag_list('TOSPLIT', [expand(elt) for elt in self.cont_cols[col2split]])
         pairs =  self.tagged.select(subinterval, ['TOSPLIT', 'LINE'])
         # group content fields by line
         rows = groupby(pairs, 1, 0)
         for row in rows:
            # process one line
            self.tagged.del_tag('TOSPLIT')
            self.tagged.tag_list('TOSPLIT',row)
            locs = self.tagged.project('TOSPLIT','LINE')
            used_hdr = set([])
            pending = []
            i = 0
            while i < len(locs):
               cont_loc = locs[i]
               found = False
               for k in range(2):
                  for j in range(len(cols)):
                     #if j in used_hdr and max(cont2hdr[i]) > j: continue
                     if k==0 and j in used_hdr: continue
                     if intersects((cont_loc, hdr_spans[j])):
                        curr_loc, next_loc = partition(cont_loc, hdr_spans, j)
                        cols[j].append(curr_loc)
                        locs = locs[:i+1] + next_loc + locs[i+1:]
                        found = True
                        used_hdr.add(j)
                        break
                  if k == 1 and not found:
                     pending.append(cont_loc)
                  elif found:
                     break
               i += 1
            cont_spans = [ut.get_list_span(col) for col in cols]
            for cont_loc in pending:
               found = False
               scores = {}
               for j in range(len(cols)):
                  if cont_spans[j] == None: continue
                  if intersects((cont_loc, cont_spans[j])):
                     found = True
                     scores[j] = intersect_len((cont_loc, cont_spans[j]))
               if found:
                  max_score = max(list(scores.values()))
                  indx = [j for j in scores.keys() if scores[j] == max_score][0]
                  cols[indx].append(cont_loc)
               else:
                  fmt = "Couldn't find column for content {}"
                  handle_error(210708, fmt.format(self.tagged.get_text_loc(cont_loc)))
               cont_spans = [ut.get_list_span(col) for col in cols]
         return cols
      col2split,hdr_spans = get_col_to_split()
      if col2split == None: return False
      cols = split_cont_col(col2split,hdr_spans )
      # replace  self.cont_cols[cc_best] with cols
      self.cont_cols = self.cont_cols[:col2split] + cols + self.cont_cols[col2split+1:]
      self.cont_cols_count = len(self.cont_cols)
      self.cont_cols_lengths = [len(col) for col in self.cont_cols]              
      return True     
   def adjust_content(self, col_count=None):
      changed = True
      while col_count != None and self.cont_cols_count < col_count and changed:
         changed = self.split_cont_cols() 
      # propagage changes 
      cont_field = [expand(loc) for col in self.cont_cols for loc in col]
      self.tagged.del_tag('cFIELD')
      self.tagged.tag_list('cFIELD', cont_field)
      locs = self.tagged.project('cFIELD', 'LINE')
      self.tagged.del_tag('pcFIELD')
      self.tagged.tag_list('pcFIELD', locs)               
      # update fields
      hdr_field = self.tagged.get_locs('hFIELD')
      self.tagged.del_tag('FIELD')
      self.tagged.tag_list('FIELD', hdr_field + cont_field)
      # update cols
      locs = self.tagged.project('FIELD', 'LINE')
      self.tagged.del_tag('pFIELD')
      self.tagged.tag_list('pFIELD', locs)
      if self.hdr_cols_count == self.cont_cols_count:
         self.cols = [self.hdr_cols[i]+self.cont_cols[i] for i in range(self.hdr_cols_count)]
      else:
         self.cols = ut.cluster(self.tagged.get_locs('pFIELD'))
      # update abs_cols
      self.abs_cols = [sorted(list(map(expand, column)), key=lambda x:x.order()) \
                       for column in self.cols]  
      self.cols_count = len(self.cols)
      self.cols_lengths = [len(col) for col in self.cols]
             
   def adjust_header(self, col_count=None):
      """
      Split columns in header if the number of columns in header 
      is smaller than the number of columns in content
      
      Merge columns in header i the number of columns in header
      is larger than the number of columns in content
      """
      split = False
      changed = True
      cont_spans = [ut.get_list_span(col) for col in self.cont_cols]
      while ((col_count == None and self.cont_cols_count > self.hdr_cols_count) or \
            (col_count != None and self.hdr_cols_count < col_count)) and \
            changed: 
         #print(self.cont_cols_count, self.hdr_cols_count)
         #print( [self.tagged.get_text_list(elt) for elt in self.hdr_cols] )
         hdr_spans  = [ut.get_list_span(col) for col in self.hdr_cols]
         # find header that covers two columns
         changed = False
         for i in range(len(cont_spans)-1):
            if intersects( ( cont_spans[i], hdr_spans[i] ) ) and \
               intersects( ( cont_spans[i+1], hdr_spans[i] ) ) :
               changed = self.split_hdr_col(i, cont_spans[i], cont_spans[i+1])
               if changed:
                  split = True
                  break  
      merged = False
      changed = True
      #print("merging...")
      while ((col_count == None and self.cont_cols_count < self.hdr_cols_count) or \
            (col_count != None and self.hdr_cols_count > col_count)) and \
            changed: 
         #print(self.cont_cols_count, self.hdr_cols_count)
         #print( [self.tagged.get_text_list(elt) for elt in self.hdr_cols] )
         hdr_spans  = [ut.get_list_span(col) for col in self.hdr_cols]
         # find 2 header that covers one column
         changed = False
         for i in range(len(hdr_spans)-1):
            if i >= len(cont_spans)-1:
               continue
            #perc_1 = intersect_len( ( cont_spans[i], hdr_spans[i] ) ) / \
            #         hdr_spans[i].length()
            perc_2 = intersect_len( ( cont_spans[i], hdr_spans[i+1] ) ) / \
                     hdr_spans[i+1].length()
            perc_3 = intersect_len( ( cont_spans[i+1], hdr_spans[i+1] ) ) / \
                     hdr_spans[i+1].length()
            if perc_2 > 0 and perc_2 > perc_3 :
               # merge headers
               merged_hdrs = self.hdr_cols[i] + self.hdr_cols[i+1]
               self.hdr_cols = self.hdr_cols[:i]+[merged_hdrs]+self.hdr_cols[i+2:]
               self.hdr_cols_count = len(self.hdr_cols)
               self.hdr_cols_lengths = [len(col) for col in self.hdr_cols]  
               merged = True
               changed = True
               break  
      # propage changes in hdr_columns to cols, abs_cols, fields, hdr_fields
      if split or merged:
         # update hdr_field related tags
         hdr_field = [expand(loc) for col in self.hdr_cols for loc in col]
         self.tagged.del_tag('hFIELD')
         self.tagged.tag_list('hFIELD', hdr_field)
         locs = self.tagged.project('hFIELD', 'LINE')
         self.tagged.del_tag('phFIELD')
         self.tagged.tag_list('phFIELD', locs)               
         cont_field = self.tagged.get_locs('cFIELD')
         # update fields
         self.tagged.del_tag('FIELD')
         self.tagged.tag_list('FIELD', hdr_field + cont_field)
         # update cols
         locs = self.tagged.project('FIELD', 'LINE')
         self.tagged.del_tag('pFIELD')
         self.tagged.tag_list('pFIELD', locs)
         if self.hdr_cols_count == self.cont_cols_count:
            self.cols = [self.hdr_cols[i]+self.cont_cols[i] for i in range(self.hdr_cols_count)]
         else:
            self.cols = ut.cluster(self.tagged.get_locs('pFIELD'))
         # update abs_cols
         self.abs_cols = [sorted(list(map(expand, column)), key=lambda x:x.order()) \
                          for column in self.cols]  
         self.cols_count = len(self.cols)
         self.cols_lengths = [len(col) for col in self.cols]
         
   def align_content_cols(self):
      """
      Aligns misaligned content cells
      """
      def get_triple(loc):
         return loc.start(), loc.end(), loc.offset
      # find content column withoud header
      for colnr in range(len(self.abs_cols)):
         col = self.abs_cols[colnr]
         col_proj = self.cols[colnr]
         # does <col> have a header?
         col_pos = [elt.order() for elt in col]
         hdr_pos = [elt.order() for elt in self.tagged.get_locs('hFIELD')]
         if set(col_pos).intersection(set(hdr_pos)):
            continue
         # <col> does not have a header, move elements to other cols
         for loc in col:
            loc_proj = [cell for cell in col_proj if expand(cell).order() == loc.order()][0]
            # get row of loc
            row = self.get_row(loc)
            row_colnrs = self.get_colnumbers(row)
            indx = [i for i in range(len(row_colnrs)) if row_colnrs[i]==colnr][0]
            newcol = None
            if indx > 0:
               prev_colnr = row_colnrs[indx-1]
               if prev_colnr < colnr-1:
                  newcol = colnr-1
            elif indx < len(row_colnrs):
               next_colnr = row_colnrs[indx+1]
               if next_colnr > colnr+1:
                  newcol = colnr+1
            if newcol == None:
               handle_error(110702, "table content columns could not be aligned")
            else:
               # move loc to column newcol
               found, i = binary_search(self.abs_cols[newcol], loc) 
               if found:
                  handle_error(110703, "location in two columns: {}".format(loc))
               self.abs_cols[newcol] = self.abs_cols[newcol][:i+1]+[loc]+\
                                       self.abs_cols[newcol][i+1:]
               # move loc_proj to column newcol
               found, i = binary_search(self.cols[newcol], loc_proj) 
               if found:
                  handle_error(110704, "location in two columns: {}".format(loc))
               self.cols[newcol] = self.cols[newcol][:i+1]+[loc_proj]+\
                                   self.cols[newcol][i+1:]
         # remove <col>, <col_proj>
         self.abs_cols = self.abs_cols[:colnr]+self.abs_cols[colnr:]
         self.cols = self.cols[:colnr]+self.cols[colnr:]
      self.cols_count = len(self.cols)
      self.cols_lengths = [len(col) for col in self.cols]
            
   def split_hdr_col(self, i, first_span, second_span):
      """
      i :           header column number
      first_span :  span of first content column covered by header i
      second_span : span of second content column covered by header i
      
      For each element in header column i, find out whether element aligns
      with first or second content column or element needs to be split into
      two strings that align with first and second content columns
      
      update self.hdr_cols, self.hdr_cols_count, and self.hdr_cols_lengths   
      """
      bef_meet = lambda pair: before(pair) or meets(pair)
      first = []
      second = []
      for hdr in self.hdr_cols[i]:
         if before((hdr, second_span)) and intersects((hdr, first_span)):
             first.append(hdr)
         elif before((first_span, hdr)) and intersects((hdr, second_span)):
             second.append(hdr)
         elif intersects((hdr, first_span)) and intersects((hdr, second_span)):
            # split hdr, it intersects with both content columns
            hdr_text = self.tagged.get_text_loc(hdr)
            #print("header",hdr_text)
            split = False
            # try splitting hdr at word boundary (e.g. ' ') 
            space_at = [i for i,ch in enumerate(hdr_text) if ch == ' ']
            # compute score for each possibility based on satisfied conditions
            satisfied = []
            for left_len in space_at:
               left_hdr = Loc(hdr.start(), hdr.start()+left_len, hdr.offset)
               right_hdr = Loc(hdr.start()+left_len+1, hdr.end(), hdr.offset)
               satisfied.append((intersects((left_hdr, first_span)),
                                 intersects((right_hdr, second_span)),
                                 bef_meet((left_hdr, second_span)),
                                 bef_meet((first_span, right_hdr)))) 
            if len(satisfied) == 0:
               # hdr is a single word, no spaces
               if len(self.hdr_cols[i]) > 1:
                  # multiple fields in header
                  # put hdr where it intersects the most
                  left_perc = intersect_len( (hdr, first_span) )/hdr.length()
                  right_perc = intersect_len( (hdr, second_span) )/hdr.length()
                  if left_perc <= right_perc:
                     left_hdr = hdr
                     right_hdr = None
                  else:
                     left_hdr = None
                     right_hdr = hdr
                  split = True
            else:
               score = [(o1+o2)*10+b1+b2 for (o1, o2, b1, b2) in satisfied]
               # get best possibility, if any
               max_score = max(score)
               if max_score > 20:
                  index = [i for i in range(len(score)) if score[i] == max_score][0]
                  left_len = space_at[index]
                  left_hdr = Loc(hdr.start(), hdr.start()+left_len, hdr.offset)
                  right_hdr = Loc(hdr.start()+left_len+1, hdr.end(), hdr.offset)
                  split = True
            if not split:
               # split in the middle between first_span and second_span
               # compute half point between cont[i] and cont[i+1]
               half_point = int((second_span.start() + first_span.end())/2)
               left_hdr = Loc(hdr.start(), half_point, hdr.offset)
               right_hdr = Loc(half_point+1, hdr.end(), hdr.offset)
            if left_hdr is not None:
               first.append(left_hdr)
            if right_hdr is not None:
               second.append(right_hdr)
         elif seq_before((first_span, hdr, second_span)):
            # header is between content columns
            # put hdr in closest column, if equidistant put in left
            left_dist = hdr.start() - first_span.end()
            right_dist = second_span.start() - hdr.end()
            if left_dist <= right_dist:
               first.append(hdr)
            else:
               second.append(hdr)
         else:
            handle_error(110705, 'Non-implemented case for split header')
      if len(first) == 0 or len(second) == 0:
          handle_error(110706, 'Could not split header')
      self.hdr_cols = self.hdr_cols[:i]+[first]+[second]+self.hdr_cols[i+1:]
      self.hdr_cols_count = len(self.hdr_cols)
      self.hdr_cols_lengths = [len(col) for col in self.hdr_cols]
      if len(first) > 0 and len(second) > 0:
         return True
      return False

   def align_cols(self):
      """
      this function fixes misalignments between header columns and content columns
      for example, consider the following table with the corresponding <cols>
         header1     header2     header3
         f1      f2          f3
         f4      f5          f6
         abs_cols = [[loc(header1),loc(f1),loc(f4)], 
                    [loc(f2),loc(f5)], 
                    [loc(header2)], 
                    [loc(f3), loc(f6)], 
                    [loc(header3)]]
         categories = [ [loc(header1),loc(header2),loc(header3)], 
                        [loc(f1), ..., loc(f6)] ]  
         where loc(f1) denotes the location of f1
      the result would be to change abs_cols to:
         [[loc(header1),loc(f1),loc(f4)], 
          [loc(header2),loc(f2),loc(f5)], 
          [loc(header3), loc(f3), loc(f6)]]  
      """
      cols = self.abs_cols
      categories = [self.tagged.get_locs('hFIELD'), self.tagged.get_locs('cFIELD')]
      cat_cols = ut.categorize(cols, categories)
      ncols = len(cols)
      # remove empty clusters
      cat_colsWE = [[ccluster for ccluster in catcluster if len(ccluster)>0] \
                   for catcluster in cat_cols]
      # compute number of columns in each category
      ncat_cols = [ len([len(ccluster) for ccluster in catcluster])\
                  for catcluster in cat_colsWE] 
      if min(ncat_cols) == max(ncat_cols) and min(ncat_cols) != ncols:
         # misaligned header and content columns
         result = []
         for c in range(len(cat_colsWE[0])):
            newcol = []
            for r in range(len(categories)):
               newcol = newcol + cat_colsWE[r][c]
            result.append(sorted(newcol, key=lambda x:x.txt_order()))
         self.abs_cols = result
        
   def get_headers_text(self, hlinecnt):
      """Get text of each header column
      
      Set data member headers to a list of text, one for each header column
      """
      self.headers = []
      # creates headers for each column
      if hlinecnt > 0:
         # split columns into header and content
         cont_field = self.tagged.get_locs('cFIELD')
         hdr_field = self.tagged.get_locs('hFIELD') 
         cat_cols = ut.categorize(self.abs_cols, [hdr_field, cont_field])   
         # select header columns
         hdr_cols = cat_cols[0]
         # join text of each header column to create a header for the column
         self.headers = [' '.join(self.tagged.get_text_list(col)) for col in hdr_cols]

   def get_rows(self):
      """Get rows from table
      
      Tag content rows in table as 'ROW_0', 'ROW_1', ...
      """
      pairs =  self.tagged.select(subinterval, ['cFIELD', 'LINE'])
      # group content fields by line
      self.rows = groupby(pairs, 1, 0)
      self.rows_count = len(self.rows)
      self.rows_lengths = [len(row) for row in self.rows]
      self.tagged.tag_lists('ROW', self.rows)
   def get_row(self, loc):
      """Finds row where Loc is
      Parameters:
         loc (Loc) -- location to search
      Returns row (Loc list) where loc is
      """
      abs_loc = expand(loc)
      for row in self.rows:
         row_pos = [elt.order() for elt in row]
         if abs_loc.order() in row_pos:
            return row
   def get_colnumbers(self, row):
      """Get column numbers of elements in row
      
      Parameters:
         row (Loc list): list of locations
      For each location in row, it finds its column number
      Returns the list of column numbers, one for each element of row 
      """
      result = []
      colnr = 0
      for loc in row:
         found = False
         for c in range(colnr,self.cols_count):
            col_loc = [elt.order() for elt in self.abs_cols[c]]
            if loc.order() in col_loc:
               colnr = c+1
               result.append(c)
               found = True
               break
         if not found:
            handle_error(110709, "Row element is not in any column")
      return result
   def untag_rows(self):
      """Remove tags 'ROW_0', 'ROW_1', from tagged"""
      for rownr in range(self.rows_count):
         self.tagged.del_tag("ROW_{}".format(rownr))
   def untag_cols(self):
      """Remove tags 'COL_0', 'COL_1', from tagged"""
      for colnr in range(self.cols_count):
         self.tagged.del_tag("COL_{}".format(colnr))
      
def get_table(text, hlinecnt, sparse=False, col_count=None):
   """Extract table from text
   
   Parameters
      text (string) -- text where table comes from
      hlinecnt (int) -- number of lines 
      sparse (bool) -- True, if there are empty cells in table (default False)
      col_count -- number of columns in table (default None, the function will
                                               determine the number of columns)
      
   The first hlinecnt lines in text that contain non-space characters are considered
   headers, the rest are considered content
   
   Returns a list of dictionaries, one for each content row. Each dictionary maps
   a header to the value in the row.
   """
   table = Table(text, hlinecnt)
   table.get_cols()
   table.get_rows()
   orig_cols_count = table.cols_count
   if col_count != None and table.cols_count != col_count:
      for i in range(2):
         if hlinecnt != 0 and col_count != table.hdr_cols_count:
            table.adjust_header(col_count)
         if col_count != table.cont_cols_count:
            table.adjust_content(col_count)
         if col_count == table.cont_cols_count and col_count == table.hdr_cols_count:
            break
         table = Table(text, hlinecnt, ut.altFIELD)
         table.get_cols()
         table.get_rows() 
   else:
      if not sparse:
         if max(table.rows_lengths) > table.hdr_cols_count:
            # check whether a single header column is covering 2 content columns 
            # if so, fix it
            table.adjust_header()
         elif max(table.rows_lengths) == table.hdr_cols_count and \
            table.cont_cols_count > table.hdr_cols_count and \
            table.cols_count > table.hdr_cols_count:
            # columns in content are misaligned
            table.align_content_cols()
      if table.hdr_cols_count == table.cont_cols_count and \
         table.cont_cols_count < table.cols_count:
         # check if header columns are misaligned with content columns
         table.align_cols()
   table.get_headers_text(hlinecnt)
   table.untag_rows()
   table.get_rows()
   # reset columns
   table.untag_cols()
   table.tagged.tag_lists('COL', table.abs_cols)
   # ------------------------------------------------------------------
   #      build table represented as a list of dictionaries 
   #      each dictionary maps a column header to a content field text
   result = []
   column_tags = ["COL_{}".format(c) for c in range(table.cols_count)]
   for rnr in range(table.rows_count):
        # get intervals in row
        locs = table.tagged.get_locs("ROW_{}".format(rnr))
        row = {}
        for loc in locs:
            # get column where loc is
            tag = table.tagged.in_tag(loc, column_tags)
            if tag == None:
               fmt = "Content field [{}] in table does not belong to any column"
               msg = fmt.format(table.tagged.get_text_loc(loc))
               handle_error(110507, msg)
            colnr = int(tag[4:]) # column nr
            if colnr >= len(table.headers) or table.headers[colnr] == '':
               if column_tags[colnr] not in row:
                  row[column_tags[colnr]] = table.tagged.get_text_loc(loc)
               else:
                  row[column_tags[colnr]] += ' ' +table.tagged.get_text_loc(loc)
            else:
               if table.headers[colnr] not in row:
                  row[table.headers[colnr]] = table.tagged.get_text_loc(loc)
               else:
                  row[table.headers[colnr]] += ' ' +table.tagged.get_text_loc(loc)
        result.append(row)
   return result



"""Module that implements location

   merge(Loc, Loc) -> Loc
   expand(Loc) -> Loc
   
   Class Loc methods:
      __init__(int, int, int)
      start() -> int
      end() -> int
      order() -> (int, int, int)
      txt_order() -> (int, int)
      length() -> int
"""
from .extracterror import handle_error

class Loc:
   """Location
   
   Constructor Parameters:
      fr (int) -- start position of interval with respect to offset
      to (int) -- end position of interval with respect to offset
      offset (int) -- position in text (default 0)
   """
   def __init__(self, fr, to, offset=0):
       if fr <= to:
           self.intrval = (fr, to) 
           self.offset = offset
       else:
           fmt = "ERROR: invalid interval {} ({}, {})"
           print(fmt.format(offset, fr, to))
           raise Exception
   def start(self):
      """Returns the start position of location"""
      return self.intrval[0]
   def end(self):
      """Returns the end position of location"""
      return self.intrval[1]
   def order(self):
      """Returns (start position, end position, offset) of location"""
      return (self.start(), self.end(), self.offset)
   def txt_order(self):
      """Returns (start position, end position) of location with respect to beginning of text"""
      return (self.offset+self.start(), self.offset+self.end())
   def length(self):
      """Returns lenght of location"""
      return (self.end() - self.start())
#------------------operations on locations--------------------------
def merge(loc1, loc2):
   """Merge two locations into one
   

   Parameters:
      loc1 (Loc) -- one of the locations to merge
      loc2 (Loc) -- the other location to merge

   Returns a new location that covers both loc1 and loc2
   It raises an exception if loc1 and loc2 do not intersect nor are contiguous
   """
   # merge overlapping intervals into one.
   # set offset of new interval to zero if input intervals have different
   # offsets
   s1, e1 = loc1.intrval
   s2, e2 = loc2.intrval
   if e1 < s2 or e2 < s1:
      handle_error(110202, "Location parameters of merge must intersect or meet")   
   offset1 = loc1.offset
   offset2 = loc2.offset
   if offset1 == offset2:
       offset = offset1
   else:
       fmt = "locations {}{}, {}{} must have same offset to be merged"
       msg= fmt.format(loc1.offset, loc1.intrval, loc2.offset, \
                        loc2.intrval)
       handle_error(110201, msg)
   return Loc(min(s1, s2), max(e1, e2), offset)
def expand(loc):
   """Converts loc to a location with offset zero
   
   Parameters:
      loc (Loc): location to be converted
      
      It translates loc into its corresponding absolute location (offset = 0)
      it returns absolute location
   """
   offset = loc.offset
   start = loc.start() + offset
   end = loc.end() + offset
   return Loc(start, end)
    



from .extracterror import handle_error

def groupby(tuplelist, groupindex, outindex):
    """
        group <tuplelist> by <groupindex>
        for each value of <groupindex>, takes tuples with that value and
        creates a list of values of <outindex> in these tuples
        for example, consider the following tuplelist where ik are locations
        and on, (sm, em) is an expansion of a location --offset,(start,end)
            tuplelist: i1, o1,(s1,e1), i3
                       i4, o3,(s2,e2), i6
                       i7, o2,(s1,e1), i8
                       i9, o4,(s2,e2), i10
            groupindex = 1
            outindex = 2
            result: [ [i3, i8], [i6, i10] ]
    """
    if len(tuplelist) == 0:
       return []
    maxindx = len(tuplelist[0])-1
    if groupindex > maxindx or groupindex < 0 or outindex > maxindx or \
       outindex < 0:
       fmt = "Parameters groupindex [{}] and outindex [{}] are out of range [0,{}]"
       msg = fmt.format(groupindex, outindex, maxindx)
       handle_error(110401, msg)   
    result = {}
    for ltuple in tuplelist:
        gloc = ltuple[groupindex]
        oloc = ltuple[outindex]
        if gloc.intrval not in result:
           result[gloc.intrval] = []
        result[gloc.intrval].append(oloc)
    return [value for key,value in result.items()]
        



"""Module for error handling

   When an error is found, funcion handle_error must be called
   
   handle_error(int, string, dict)

"""
class ExtractError(Exception):  
   def __init__(self, code, msg):
      self.code = code
      self.msg = msg
      
def handle_error(code, msg, error_loc=None):
   """Function to handle errors
   
   Parameters:
      code (int) -- error code consisting of 6 digits: tammdd
          t is type of error: 
             1 for serious, 
             2 for warning
          a is the application
                1 for qante
          mm is module where it occurred (should be unique within application):
                for tagging
                   01 is tagger
                   02 for loc
                   03 for loctuple
                   04 for loctuplelist
                   05 for utilities
                   06 for query
                   07 for table
                   09 for loclist
          dd is error code within module
      msg (string) -- error message
      error_loc (dict) -- unused                                   
   """  
   if code < 100000 or code > 999999:
      pref = "ERROR: invalid error code and "
   else:
      t = int(code/100000) 
      if t == 1:
         pref = "ERROR[{}]: ".format(code)
         raise ExtractError(code, pref+msg)
      elif t == 2:
         pref = "Warning[{}]: ".format(code)
      else:
         pref = "ERROR: invalid error code and " 
      print(pref + msg)





"""Module that implements relations and functions on location tuples

Allen's relations:
   before(Loc tuple) -> boolean
   meets(Loc tuple) -> boolean
   overlaps(Loc tuple) -> boolean
   starts(Loc tuple) -> boolean
   during(Loc tuple) -> boolean
   finishes(Loc tuple) -> boolean
   equal(Loc tuple) -> boolean
Other relations:
   seq_meets(Loc tuple) -> boolean
   seq_before(Loc tuple) -> boolean
   seq_before_meets(Loc tuple) -> boolean
   subinterval(Loc tuple) -> boolean
   intersects(Loc tuple) -> boolean
   dijoint(Loc tuple) -> boolean
   
This module also defines functions that map location tuples into a location or a tuple location
   closed_span(Loc tuple) -> Loc
   open_right_span(Loc tuple) -> Loc
   open_left_span(Loc tuple) -> Loc
   open_span(Loc tuple) -> Loc
   last(Loc tuple) -> Loc
   first(Loc tuple) -> Loc
   in_between(Loc tuple) -> Loc tuple

"""
from .loc import Loc
from .extracterror import handle_error

# ---------------allen's interval relations------------------------
# each pair consists of two locations, 
# each location consists of an offset and an interval
# allen's relations are applied to the intervals
def before(pair):
   """Allen's before relation
   
   Parameters:
      pair (Loc tuple) -- tuple with two locations
   
   Locations are represented by offset, start, end. This function returns True
   if first location appears before second location by considering start and
   end only, disregarding their offset. Otherwise, it returns False

   """
   if len(pair) != 2:
      handle_error(110306, "before requires a 2-tuple as input parameter")
   s1, e1 = pair[0].intrval
   s2, e2 = pair[1].intrval
   return e1 < s2
def meets(pair):
   """Allen's meets relation
   
   Parameters:
      pair (Loc tuple) -- tuple with two locations
   
   Locations are represented by offset, start, end. This function returns True
   if first location meets second location by considering start and
   end only, disregarding their offset. Otherwise, it returns False

   """
   if len(pair) != 2:
      handle_error(110306, "meets requires a 2-tuple as input parameter")
   s1, e1 = pair[0].intrval
   s2, e2 = pair[1].intrval
   return e1 == s2 
def overlaps(pair):
   """Allen's overlaps relation
   
   Parameters:
      pair (Loc tuple) -- tuple with two locations
   
   Locations are represented by offset, start, end. This function returns True
   if first location overlaps second location by considering start and
   end only, disregarding their offset. Otherwise, it returns False

   """
   if len(pair) != 2:
      handle_error(110306, "overlaps requires a 2-tuple as input parameter")
   s1, e1 = pair[0].intrval
   s2, e2 = pair[1].intrval
   return s1 < s2 and s2 < e1 and e1 < e2
def starts(pair):
   """Allen's starts relation
   
   Parameters:
      pair (Loc tuple) -- tuple with two locations
   
   Locations are represented by offset, start, end. This function returns True
   if first location starts second location by considering start and
   end only, disregarding their offset. Otherwise, it returns False

   """
   if len(pair) != 2:
      handle_error(110306, "starts requires a 2-tuple as input parameter")
   s1, e1 = pair[0].intrval
   s2, e2 = pair[1].intrval
   return s1 == s2 and e1 < e2 
def during(pair):
   """Allen's during relation
   
   Parameters:
      pair (Loc tuple) -- tuple with two locations
   
   Locations are represented by offset, start, end. This function returns True
   if first occurs during second location by considering start and
   end only, disregarding their offset. Otherwise, it returns False

   """
   if len(pair) != 2:
      handle_error(110306, "during requires a 2-tuple as input parameter")
   s1, e1 = pair[0].intrval
   s2, e2 = pair[1].intrval
   return s2 < s1 and e1 < e2
def finishes(pair):
   """Allen's finishes relation
   
   Parameters:
      pair (Loc tuple) -- tuple with two locations
   
   Locations are represented by offset, start, end. This function returns True
   if first location finishes second location by considering start and
   end only, disregarding their offset. Otherwise, it returns False

   """
   if len(pair) != 2:
      handle_error(110306, "finishes requires a 2-tuple as input parameter")
   s1, e1 = pair[0].intrval
   s2, e2 = pair[1].intrval
   return s2 < s1 and e1 == e2
def equal(pair):
   """Allen's equal relation
   
   Parameters:
      pair (Loc tuple) -- tuple with two locations
   
   Locations are represented by offset, start, end. This function returns True
   if first location is equal to second location by considering start and
   end only, disregarding their offset. Otherwise, it returns False

   """
   if len(pair) != 2:
      handle_error(110306, "equal requires a 2-tuple as input parameter")
   s1, e1 = pair[0].intrval
   s2, e2 = pair[1].intrval
   return s1 == s2 and e1 == e2 
# --------------additional interval relations-------------------
def seq_meets(ltuple):
   """This relation extends meets to more than two arguments
      
      Parameters:
         ltuple (Loc tuple): tuple of locations
      
      If ltuple = (l1, ..., ln), it returns meets((l1,l2)) and meets((l2,l3))
      and ... and meets((ln-1,ln))
      
   """
   for i in range(len(ltuple)-1):
       if not meets( [ltuple[i], ltuple[i+1]] ):
           return False
   return True
def seq_before(ltuple):
   """This relation extends before to more than two arguments
      
      Parameters:
         ltuple (Loc tuple): tuple of locations
      
      If ltuple = (l1, ..., ln), it returns before((l1,l2)) and before((l2,l3))
      and ... and before((ln-1,ln))
   """
   for i in range(len(ltuple)-1):
       if not before( [ltuple[i], ltuple[i+1]] ):
           return False
   return True
def seq_before_meets(ltuple):
   """This relation combines before and meets with a disjunction
   
   Parameters:
      ltuple (Loc tuple): tuple of locations
      
   If ltuple = (loc1, ..., locn), it returns (before((l1,l2)) or meets((l1,l2)))
   and (before((l2,l3)) or meets((l2,l3))) and ... and (before((ln-1,ln)) or meets((ln-1,ln)))

   """
   for i in range(len(ltuple)-1):
       if not before( [ltuple[i], ltuple[i+1]] ) and not meets( [ltuple[i], ltuple[i+1]] ):
           return False
   return True
def subinterval(pair):
   """This relation tests whether first interval is a subset or equal to second interval
   
   Parameters:
      pair (Loc tuple): a pair of locations
      
   Locations are represented by offset, start, end. This function returns True
   if interval of first location intersects interval of second location.
   Intervals are determined by start and end of locations, disregarding offset. 
   """
   if len(pair) != 2:
      handle_error(110306, "subinterval requires a 2-tuple as input parameter")
   return starts(pair) or during(pair) or finishes(pair) or equal(pair)
def intersects(pair): 
   """This relation tests whether first interval intersects with second interval
   
   Parameters:
      pair (Loc tuple): a pair of locations
   
   Locations are represented by offset, start, end. This function returns True
   if interval of first location intersects interval of second location.
   Intervals are determined by start and end of locations, disregarding offset. 
   """
   if len(pair) != 2:
      handle_error(110306, "intersects requires a 2-tuple as input parameter")
   rpair = (pair[1], pair[0]) # reverse pair
   ret = overlaps(pair) or starts(pair) or during(pair) or finishes(pair) or \
         overlaps(rpair) or starts(rpair) or during(rpair) or finishes(rpair) \
         or equal(pair)
   return ret
def disjoint(pair):
   """This relation tests whether first interval is dijoint with second interval
   
   Parameters:
      pair (Loc tuple): a pair of locations
   
   Locations are represented by offset, start, end. This function returns True
   if interval of first location does not intersect interval of second location.
   Intervals are determined by start and end of locations, disregarding offset. 
   """
   if len(pair) != 2:
      handle_error(110306, "disjoint requires a 2-tuple as input parameter")
   return not intersects(pair)
def intersect_len(pair):
   s1,e1 = pair[0].intrval
   s2,e2 = pair[1].intrval
   if s1 < s2 and s2 < e1:
      return min(e1,e2)-s2
   elif s2 < s1 and s1 < e2:
      return min(e1,e2)-s1
   elif s1 == s2:
      return min(e2-s2, e1-s1)
   else:
      return 0 
def begins(pair):
   return starts(pair) or equal(pair)
def closed_span(pair):
   """Returns a location that spans from beginning of first location to end of second location
   
   Parameters:
      pair (Loc tuple) -- pair of locations
       
   It raises an exception if two locations in pair have different offset
   """
   if len(pair) != 2:
      handle_error(110306, "closed_span requires a 2-tuple as input parameter")
   offset = pair[0].offset
   if offset != pair[1].offset:   
      msg = "Function closed_span requires locations to have same offset"
      handle_error(110301, msg)
   start = pair[0].start()
   end = pair[1].end()
   
   return Loc(start, end, offset)
def last(ltuple):
   """Returns last location in ltuple
   
   Parameters:
      ltuple (Loc tuple) -- tuple of locations
   """
   n = len(ltuple)
   return ltuple[n-1]
def first(ltuple):
   """Returns first location in ltuple 
   
   Parameters:
      ltuple (Loc tuple) -- tuple of locations
   """
   return ltuple[0]
def open_right_span(pair):
   """Returns a location that spans from start of first location to start of second location
      
      Parameters:
         pair (Loc tuple) -- pair of locations
          
      It raises an exception if two locations in pair have different offset
   """
   if len(pair) != 2:
      handle_error(110306, "open_right_span requires a 2-tuple as input parameter")
   offset = pair[0].offset  
   if pair[1].offset != offset:
      msg = "Function open_right_span requires locations to have same offset"
      handle_error(110302, msg)
   fr = pair[0].start()
   to = pair[1].start()
   return Loc(fr,to,offset)
def open_left_span(pair):
   """Returns a location that spans from end of first location to end of second location
      
      Parameters:
         pair (Loc tuple) -- pair of locations
          
      It raises an exception if two locations in pair have different offset
   """
   if len(pair) != 2:
      handle_error(110306, "open_left_span requires a 2-tuple as input parameter")
   offset = pair[0].offset
   if pair[1].offset != offset:
      msg = "Function open_right_span requires locations to have same offset"
      handle_error(110303, msg)
   fr = pair[0].end()
   to = pair[1].end()
   return Loc(fr,to,offset)
def open_span(pair):
   """Returns a location that spans from end of first location to start of second location
      
      Parameters:
         pair (Loc tuple) -- pair of locations
          
      It raises an exception if two locations in pair have different offset
   """
   if len(pair) != 2:
      handle_error(110306, "open_span requires a 2-tuple as input parameter")
   offset = pair[0].offset
   if pair[1].offset != offset:
      msg = "Function open_span requires locations to have same offset"
      handle_error(110304, msg)
   fr = pair[0].end()
   to = pair[1].start()
   return Loc(fr,to,offset)
def in_between(ltuple):
   """Returns the locations between the locations in ltuple
   
   Parameters:
      ltuple (Loc tuple) -- tuple of locations (loc1, loc2, ..., locn)
      
   Returns tuple of locations: (rloc1, rloc2, ..., rlocn-1) where rloci is the 
   location between loci and loci+1
   It raises an exception if locations in tuple do not have the same offset

   """
   offset = ltuple[0].offset
   fr = ltuple[0].end()
   res = []
   for loc in ltuple[1:]:
      if loc.offset != offset:
         msg = "Function in_between requires locations to have same offset"
         handle_error(110305, msg)
      to = loc.start()
      res.append(Loc(fr,to,offset))
      fr = loc.end()
   return tuple(res)
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Module that processes queries on tagged objects
   
   Class Query methods:
      __init__(string/literal list, string, Tagger object, int list, boolean)
      execute() -> list of Loc tuples
      UDP(string, lambda)

"""
import regex as re

from .extracterror import handle_error
from .loctuple import subinterval, seq_before, meets, starts
from .loctuple import before, seq_meets, equal, intersects, disjoint
from .loctuple import overlaps, seq_before_meets

def rm_dups(tuples):
   included = set([])
   sresult = []
   for r in tuples:
      if r in included: continue
      sresult.append(r)
      included.add(r)
   return sresult
def merge_results(prev, new):
   """
      prev : list of schema-tuples pairs
      new :  list of schema-tuples pairs

      updates prev to include new
      if prev already includes tuples for a schema in new, it add new tuples

   """
   for schema, tuples in new:
      found = False
      for i, (osch, otup) in enumerate(prev):
         if schema == osch:
            found = True
            prev[i] = (osch, rm_dups(otup + tuples))
            break
      if not found:
          prev.append( (schema, tuples))
def natural_inner_join(sch_tuples1, sch_tuples2):
   """
      applies natural inner join to tuples1 and tuples2
      schemas of tuples1 and tuples2 must intersect

      sch_tuples1 : pair (sch1,tuples1) where sch1 is a list of integers
                    denoting column number, and tuples1 is a list of location
                    tuples that follow sch1
      sch_tuples2 : pair (sch2,tuples2) where sch2 is a list of integers
                    denoting column number, and tuples2 is a list of location
                    tuples that follow sch2

      Returns a pair (sch, tuples) where sch is the sorted union of sch1 and sch2 and
                    tuples is the result of applying natural join to tuples1 and
                    tuples2 and follows schema sch
   """
   sch1, tuples1 = sch_tuples1
   sch2, tuples2 = sch_tuples2
   overlap = list(set(sch1).intersection(set(sch2)))
   if len(overlap) == 0:
      handle_error(110601, 'schemas in natural_inner_join must intersect')
   if len(sch1) != len(set(sch1)) or len(sch2) != len(set(sch2)):
      handle_error(110602, 'schema has duplicate column number')
   # overlap column number to sch1 index
   over2sch1 = {}
   for o in overlap:
      over2sch1[o] =[i for i in range(len(sch1)) if sch1[i] == o][0]
   # overlap column number to sch2 index
   over2sch2 = {}
   for o in overlap:
      over2sch2[o] =[i for i in range(len(sch2)) if sch2[i] == o][0]
   # sort tuples by overlapping columns
   stuples1 = sorted(tuples1, key=lambda t: tuple( [t[over2sch1[o]].order() for o in overlap] ))
   stuples2 = sorted(tuples2, key=lambda t: tuple( [t[over2sch2[o]].order() for o in overlap] ))
   t1 = 0
   t2 = 0
   joinres = []
   joinsch = sch1 + [col for col in sch2 if col not in overlap]
   # compute join
   while t1 < len(stuples1) and t2 < len(stuples2):
      tuple1 = stuples1[t1]
      tuple2 = stuples2[t2]
      key1 = tuple([tuple1[over2sch1[o]].order() for o in overlap])
      key2 = tuple([tuple2[over2sch2[o]].order() for o in overlap])
      if key1 < key2:
         t1 += 1
      elif key1 > key2:
         t2 += 1
      else:
         i = t1
         while i < len(stuples1):
            j = t2
            while j < len(stuples2):
               tuple1 = stuples1[i]
               tuple2 = stuples2[j]
               k1 = tuple([tuple1[over2sch1[o]].order() for o in overlap])
               k2 = tuple([tuple2[over2sch2[o]].order() for o in overlap])
               if k1 == key1 and k2 == key2:
                  l2 = [tuple2[col] for col in range(len(sch2)) if sch2[col] not in overlap]   
                  joinres.append(tuple(list(tuple1)+l2))
                  j += 1
               else:
                  break
            i += 1
         t1 += 1
         t2 += 1
   # sort columns in joinres so that first column is the one with the lowest
   # column number, then next corresponds to next column number, and so forth
   scols = sorted(joinsch)
   colsmap = []
   for col in scols:
      indx = [i for i in range(len(joinsch)) if joinsch[i] == col][0]
      colsmap.append(indx)
   stuples =[] 
   for t in joinres:
      new_t = [t[indx] for indx in colsmap]
      stuples.append(tuple(new_t))
   return (scols, stuples)
def cartesian_prod(tuples, ptuples, schema):
   """
      tuples :      list of location tuples
      ptuples :     list of location tuples
      schema :      dictionary 
         {'pred':   [c1, ..., cp] : column numbers in ptuples
          'tuples': [d1, ..., dt] : column numbers in tuples
          'tags':   unused

      Computes cartesian product of tuples x ptuples and sorts
      columns in cartesian product by column number
      returns column numbers of result, sorted cartesian product
   """ 
   spred = schema['pred']
   stuples = schema['tuples']
   # verify that schema['pred'] and schema['tuples'] are disjoint
   if len(set(spred).intersection(set(stuples))) != 0:
      fmt = 'schemas of cartesian product operands must be disjoint: {} {}'
      handle_error(110603, fmt.format(spred, stuples))
   # column numbers of result
   soutput = sorted(spred+stuples)
   # mapping column number in result to where it comes from
   col2src = {}
   for col in soutput:
      if col in spred:
         src = 'p'
         indx = [i for i in range(len(spred)) if spred[i] == col][0]
      else:
         src = 't'
         indx = [i for i in range(len(stuples)) if stuples[i] == col][0]
      col2src[col] = (src, indx)  
   result = []
   for ttuple in tuples:
      for ptuple in ptuples:
         rtuple = []
         for col in soutput:
            src,indx = col2src[col]
            if src == 'p':
               rtuple.append(ptuple[indx])
            else:
               rtuple.append(ttuple[indx])
         result.append(tuple(rtuple))
   return soutput,result  
def project(prev_result, oschema):
    """
        prev_result : a schema-tuples pair, where schema is a list of columns
                      and tuples is a list of tuples following the schema
        oschema :     list of colums to project prev_result on
        
        takes tuples in prev_result and project on columns in oschema
        
        Returns a pair new_sch, new_tuples
           new_sch :    list of columns
           new_tuples : list of tuples
    """
    prev_sch, prev_tuples = prev_result
    new_sch = sorted(list(set(prev_sch).intersection(set(oschema))))
    if new_sch == prev_sch:
        return prev_result
    indices = [i for i in range(len(prev_sch)) if prev_sch[i] in new_sch]
    new_tuples = []
    for ptuple in prev_tuples:
        ltuple = list(ptuple)
        new_tuple = tuple( [ltuple[i] for i in range(len(ltuple)) if i in indices] )
        new_tuples.append(new_tuple)
    return new_sch, new_tuples
class Node:
   def __init__(self, op):
      self.op = op          # operator: predicate or booleans AND, OR
      self.result = []      # result so far, list of (schema,[(loc1, ..., locn),...])
      self.children = []    # children, if any
      self.params = []      # parameters as list of column numbers, if leaf
      self.ecount = None    # number of tuples to consider in computation
      self.done = False     # whether computation of node was completed
      self.processed =[]
   def print_node(self, pref, tm, fd=None):
      if fd == None:
         print(pref,self.op)
         if len(self.result) > 0:
            print(pref+"Current result")
            for schema, tuples in self.result:
               print(pref+"schema: {}, {} tuples".format(schema, len(tuples)))
               tm.display_tuples(tuples)
               if self.result not in self.processed:
                  self.processed.append(self.result)
         if len(self.children) == 0:
            print(pref+"leaf")
         if self.done:
            print(pref+"DONE")
         else:
            print(pref+"params: ",self.params)
            print(pref+"estimate of computation count: ",self.ecount)
            print(pref+"pending")
      else:
         line = ["{}{}---params: {}".format(pref,self.op,self.params)]
         if self.done:
            for schema, tuples in self.result:
               line.append(pref+"schema: {}, actual count {} tuples".format(schema, len(tuples)))
         else:
            for schema, tuples in self.result:
               line.append(pref+"schema: {}, current count {} tuples".format(schema, len(tuples)))
            line.append(pref+"estimate count {} tuples".format(self.ecount))
         for l in line: 
            fd.write(l)
            fd.write('\n')

class Query:
   """Module for querying tagged text
      
   Constructor Parameters:
      tags (string/literal list) -- list of tags or literals used by query
      query (string) -- query to be executed. it must follow this syntax:
         query ::= (query)
         query ::= conj
         query ::= disj
         query ::= term
         conj ::= term and query
         disj ::= term or query
         term ::= pred ( params )
         term ::= dist (int, int) relop int
         relop ::= < | > | = | <= | >= | !=
         params ::= int , params
         params ::= int
         pred is any predicate in self.PREDS and int is an integer
      tagger (Tagger Object) -- tagged text to apply query
      project (int list) -- list of indices in tags to include in result (default [])
      log_on (boolean) -- whether to log query execution on file query_exec_log.txt (default False)
   """
   FSAdist = {0:[('[(]',1,'token')],
          1:[('[0-9]+', 2, 'token')],
          2:[(',',3,'token')],
          3:[('[0-9]+', 4, 'token')],
          4:[('[)]',5,'token')],
          5:[('[<>=!]', 6, None)],
          6:[('=',7,'combine2'), ('[0-9]+', 8, '2tokens')],
          7:[('[0-9]+', 8, 'token')]}
   def __init__(self, tags, query, tagger, project = [], log_on=False):
      self.PREDS = {'subinterval': subinterval, 
               'seq_before': seq_before,
               'meets': meets,
               'starts': starts,
               'before': before,
               'equal': equal,
               'intersects': intersects,
               'disjoint': disjoint,
               'overlaps': overlaps,
               'seq_before_meets': seq_before_meets,
               'seq_meets': seq_meets}
      if log_on:
         self.fd = open("query_exec_log.txt",'a')
      else:
         self.fd = None
      self.root = None
      self.tagger = tagger
      # subinterval(1, 0) and subinterval(4, 3) and seq_before(0, 2, 3)
      self.query = query 
      self.tokens = []
      # tags considered in this query and number of locations associated with each
      self.qtags = {i:tags[i] for i in range(len(tags))}
      self.count = {i:len(tagger.get_locs(tags[i])) for i in range(len(tags))}
      empty_tags = [self.qtags[i] for i in self.count.keys() if self.count[i] == 0]
      if len(empty_tags) != 0:
          handle_error(210604, 'Tags in query are empty: {}'.format(empty_tags))
      # list of estimated counts of leaves that have not been computed
      self.ecounts = []
      self.project = project
      if len(set(project).difference(set(range(len(self.qtags))))) != 0:
          fmt = "Invalid project columns in query: {}"
          handle_error(110605, fmt.format(project))
   def __del__(self):
      if self.fd != None:
         self.fd.close()
   def UDP(self, pred_name, pred_function):
      """User Defined Predicate
      
      Parameters:
         pred_name (string) -- name of predicate
         pred_function (lambda) -- boolean function to apply when predicate is invoked. 
         
      Defines a new predicate to be included in queries
      """
      self.PREDS[pred_name] = pred_function
   def tokenize(self):
      def parse_dist(i, tokens):
         """
             Verify that next tokens follow this pattern: ( n1 , n2 ) op n3
             where op is a single token (<,>,=) or two tokens(< =, > =, ! =)
             If so, return the following list: ( n1 , n2 ) op n3 where op
             is a single token (tokens < =, > =, ! = are comvined into <=, >=, != )
         """
         token_list = []
         state = 0
         for j in range(i,len(tokens)):
            if state == 8:
               return (j, token_list)
            found = False
            for (pattern, state, op) in Query.FSAdist[state]:
               m = re.search('^{}$'.format(pattern), tokens[j])
               if m != None: 
                  found = True                 
                  break
            if found:
               if op == 'token':
                  token_list.append(tokens[j])
               elif op == 'combine2':
                  token_list.append(tokens[j-1]+tokens[j])
               elif op == '2tokens':
                  token_list.append(tokens[j-1])
                  token_list.append(tokens[j])
            else:
               fmt = 'Invalid expression: {}'
               handle_error(110606, fmt.format(' '.join(tokens[i-1:j+1])))
         if state == 8:
            return (j+1, token_list)
         else:
            fmt = 'Invalid expression: {}'
            handle_error(110607, fmt.format(' '.join(self.tokens[i-1:j+1])))
      if len(self.tokens) > 0:
         fmt = "Query was already tokenized: {}"
         handle_error(210608, fmt.format(self.tokens))
         
      word_patt = '[\w]+'
      delim_patt = '[)(,><=!]'
      pattern = "({})|({})".format(word_patt, delim_patt)
      tokens = []
      for m in re.finditer(pattern, self.query):
         tokens.append(m.group())
      # --------
      # replace dist(n1, n2) op n3, where op in [<. >, =, <=, >=, !=]
      # with dist_pred(_n(n1, n2) where
      # dist_pred_n = lambda t: t[0].offset == t[1].offset and 
      #                         t[1].start() >= t[0].end() and 
      #                         t[1].start()-t[0].end()) op n3
      # --------
      self.tokens = []
      i = 0
      while i < len(tokens):
         if tokens[i] == 'dist':
            n = len(self.PREDS.keys())
            fn_str = 'dist_pred_{}'.format(n)
            self.tokens.append(fn_str)
            i,token_list = parse_dist(i+1, tokens)
            self.tokens = self.tokens + token_list[:-2]
            fmt = "lambda t: (t[0].offset == t[1].offset) and " + \
                  "(t[1].start() >= t[0].end()) and " + \
                  "(t[1].start()-t[0].end()) {} {}"
            lambda_fn = fmt.format( token_list[-2], token_list[-1])
            self.PREDS[fn_str] = eval( lambda_fn )
         else:
            self.tokens.append(tokens[i])
            i += 1
            
   def parse(self):
      """
         Processes list of tokens in self.tokens to compute a parse tree
         It sets self.root to the root of the parse tree
      """
      def is_bop(token):
         # is token a boolean operator?
         return token == 'and' or token == 'or'
      def get_pred_params(i):
         """
            i: next index in self.tokens after predicate name
            
            parses "(c1, .., cn)" after predicate name

            Returns pair (params, i) where
               params : list [c1, ..., cn]
               i: index in self.tokens of closing parenthesis ')' after cn
         """
         nxt = 'open_paren'
         params = []
         while i < len(self.tokens) and self.tokens[i] != ')':
            if (nxt == 'num' and re.search('^[0-9]+$', self.tokens[i]) == None) or \
               (nxt == 'comma' and self.tokens[i] != ',') or \
               (nxt == 'open_paren' and self.tokens[i] != '('):
               fmt = "Invalid syntax in query: {} - Error while parsing parameters at: {}"
               handle_error(110609, fmt.format(self.query, self.tokens[i]))
            if nxt == 'num':
               val = int(self.tokens[i])
               if val < 0 or val >= len(self.qtags):
                  fmt = 'Parameter {} of predicate out of range. In query: {}'
                  handle_error(110610, fmt.format(val, self.query))
               params.append(val)
               nxt = 'comma'
            else:
               nxt = 'num'
            i += 1
         return params, i
      def get_subtree(bool_ops, preds, token):
         """
            bool_ops: stack of boolean operators 'and' and 'or'
            preds: stack of predicates or roots of subtrees
            token: token being processed and is
            used in the error message if an error is found
            
            builds subtree with top of bool_ops as the parent and 
            top two elements of preds as the children
            
            pushes resulting msubtree into preds stack
         """
         op = bool_ops.pop()
         if len(preds) < 2:
            fmt = "Invalid syntax in query: {} - {}"
            handle_error(110611,fmt.format(self.query, token))
         child1 = preds.pop()
         child2 = preds.pop()
         node = Node(op)
         node.children = [child1, child2]
         preds.append(node)
      def get_tree(start_index, open_paren):
         """
            start_index : index in self.tokens where the parsing begins
            open_paren: boolean indicating this function was called as a
            result of finding an open parenthesis
            
            builds a parse tree from tokens starting at index star_index
            if open_paren is True, then stops when the closing parenthesis
            is found, otherwise it stops at the end of self.tokens
       
            Returns the root of the tree
         """
         i = start_index
         bool_ops = []
         preds = []
         nxt_type = 'pred'
         while i < len(self.tokens):
            token = self.tokens[i]
            # check for parsing errors
            if (nxt_type == 'pred' and is_bop(token)) or \
               (nxt_type == 'bool_op' and not is_bop(token) and token != ')') or \
               (token == ')' and not open_paren):
               fmt = "Invalid syntax in query: {} - {}"
               handle_error(110612,fmt.format(self.query, token))
            # pushes a boolean operator (and/or) into bool_ops stack
            if is_bop(token):
               bool_ops.append(token)
               nxt_type = 'pred'
               i += 1
            # parses subexpression starting at '(' by making a recursive call
            elif token == '(':
               root, index = get_tree(i+1, True)
               preds.append(root)
               nxt_type = 'bool_op'
               i = index
            # finishes parsing of subexpression ending at ')'
            elif  token == ')':
               while len(bool_ops) > 0:
                  get_subtree(bool_ops, preds, token)
               return preds.pop(), i+1
            # creates node with predicate and its parameters
            # pushes node into preds stack
            else:
               # get predicate
               if token not in self.PREDS:
                  fmt = "Invalid predicate in query; {}"
                  handle_error(110613, fmt.format(token))
               function = self.PREDS[token]
               node = Node(function)
               node.params, i = get_pred_params(i+1)
               preds.append(node)
               # create subtree if top of bool_ops is 'and' with 'and' as parent
               # and top two elements in preds stack as children
               # pop 'and' and top two elements
               # push subtree into preds stack
               if len(bool_ops) > 0 and bool_ops[-1] == 'and':
                  get_subtree(bool_ops, preds, token)
               nxt_type = 'bool_op'
               i += 1
         while len(bool_ops) > 0:
            get_subtree(bool_ops, preds, "")
         if len(preds) == 0:
            fmt = "Invalid query; {}"
            handle_error(110614, fmt.format(self.query))
         return preds.pop()
      self.root = get_tree(0, False)
   def flatten_tree(self):
      """
         combine contiguous 'and' or 'or' nodes into a single node
         ex:
            Original Tree:
                           and
                            |
                        --------
                        |      |
                       and     or
                        |      |
                     ------  -----
                     |    |  |   |
                     x    y  z   w
          Transformed Tree:
                          and
                           |
                     -----------
                     |   |     |   
                     x   y     or
                               |
                             -----
                             |   |
                             z   w
                  
      """
      def depth_first(ptr):
         if ptr.op == 'and' or ptr.op == 'or':
            to_rm = []
            bool_op = ptr.op
            for indx, child in enumerate(ptr.children):
               if child.op == bool_op:
                  to_rm.append(indx)
                  for grandchild in child.children:
                     ptr.children.append(grandchild)
            for r in range(len(to_rm)-1, -1, -1):
               ptr.children = ptr.children[:to_rm[r]]+ptr.children[to_rm[r]+1:]
            for child in ptr.children:
               depth_first(child)
      depth_first(self.root)
   def find_ecount(self, ecount):
     lo = 0
     hi = len(self.ecounts)-1
     while lo <= hi:
        mid = int( (lo+hi) / 2)
        if self.ecounts[mid] == ecount:
           return (True, mid)
        elif self.ecounts[mid] < ecount:
           lo = mid+1
        else:
           hi = mid-1
     return (False, hi)
   def delete_ecount(self, ecount):
      found, indx = self.find_ecount(ecount) 
      if not found:
         fmt = 'Trying to delete unexistent ecount {} in {}'
         handle_error(110615,fmt.format(ecount), self.ecounts)
      self.ecounts = self.ecounts[:indx]+self.ecounts[indx+1:]
   def update_ecount(self, oldcnt, newcnt):
      # delete oldcnt
      self.delete_ecount(oldcnt)
      # insert newcnt
      found, indx = self.find_ecount(newcnt) 
      self.ecounts = self.ecounts[:indx+1]+[newcnt]+self.ecounts[indx+1:] 
   def add_ecount(self, newcnt):
      found, indx = self.find_ecount(newcnt) 
      self.ecounts = self.ecounts[:indx+1]+[newcnt]+self.ecounts[indx+1:]           
   def update_tree(self):
      """
         traverse tree to update results of 'and'/'or' nodes if computation of
         their children was completed. 
         it traverses the tree depth-first and updates the nodes when returning
         to the node after visiting all its descendants
         moves up the results to update 'and' nodes results 
         update ecount of leaves
      """
      def update_node(ptr, parent):
         """
            ptr :    inner node
            parent : parent of <ptr>
            
            check whether all children of <ptr> have completed their computation
            if so, update <ptr> and <parent> states
         """
         # check whether all children are done
         ptr.done = sum([1 for child in ptr.children if not child.done]) == 0
         if ptr.done:
            # this node is done, update and stop recursion
            # move result one level up
            if parent != None:
               if parent.op == 'and':
                  if len(parent.result) == 0:
                     parent.result = ptr.result
                  elif len(ptr.result) == 0:
                     msg = "Result of conjunction is empty, it should include a schema"
                     handle_error(110616, msg )
                  else:
                     # apply join AQUI -- parent.result = naturaljoin(parent.result,ptr.result)
                     new_parent_result =[]
                     for s1,t1 in parent.result:
                        for s2,t2 in ptr.result:
                           schema_inter = set(s1).intersection(set(s2))
                           if len(schema_inter) > 0:
                              schema,tuples = natural_inner_join((s1,t1), (s2,t2))
                           else:
                              schema,tuples = cartesian_prod(t1, t2, schema={'tuples':s1,'pred':s2})
                           new_schema = True
                           for i, pair in enumerate(new_parent_result):
                              s,t = pair
                              if schema == s:
                                 new_parent_result[i] = (s, rm_dups(t + tuples))
                                 new_schema = False
                                 break
                           if new_schema:
                              new_parent_result.append((schema,tuples))  
                     parent.result = new_parent_result      
               else:
                  # parent is 'or', append result
                  merge_results(parent.result, ptr.result)
      def depth_first(ptr, parent, prev_res):
         if len(ptr.children) == 0:
            # a leaf
            if not ptr.done:
               prev_ecount = ptr.ecount
               if prev_res == None:
                  ptr.ecount = 1
                  for col in ptr.params:
                     ptr.ecount = ptr.ecount * self.count[col]
               else:
                  ptr.ecount = 0
                  for cols, tuples in prev_res:
                     extra_cols = [col for col in ptr.params if col not in cols]
                     if extra_cols == ptr.params:
                        cnt = 1
                     else:
                        cnt = len(tuples)
                     for col in extra_cols:
                        cnt = cnt * self.count[col]
                     ptr.ecount += cnt
               # update self.ecounts
               if prev_ecount == None:
                  self.add_ecount(ptr.ecount)
               else:
                  self.update_ecount(prev_ecount, ptr.ecount)
         else:
            # inner node
            if not ptr.done:
               if ptr.op == 'and' and len(ptr.result) > 0:
                  # and's current results move down
                  prev_res = ptr.result
               for child in ptr.children:
                  depth_first(child, ptr, prev_res)
               update_node(ptr, parent)
      depth_first(self.root, None, None)
        
   def print_tree(self):
      def depth_first(ptr, pref):
         ptr.print_node(pref,self.tagger, self.fd)
         if len(ptr.children) > 0:
            pref = pref + "---"
            for child in ptr.children:
               depth_first(child, pref)
      depth_first(self.root,"")
      if self.fd == None:
         print("\n\n")
      else:
         self.fd.write('\n\n')
         self.fd.flush()
   def print_result(self, result):
      for cols,tuples in result:
         print('===')
         for t in tuples:
            print('---')
            for i in range(len(t)):
               f = t[i]
               print(self.tagger.get_text_loc(f), '\n...')
            
   def compute_leaf(self):
      """
         Traverse tree depth-first until it finds leaf with lowest value of ecount
         When going down the tree and finding an 'and' node, take the result of
         'and' node and push down as previous result. when a new 'and' node is
         found the previous result is updated with the result of new 'and' node
         When leaf with lowest value is found, it uses previous result to compute the
         result of the leaf and updates the result of the leaf parent.
         if leaf parent is 'and', it overwrites its result with computed result
         if leaf parent is 'or', it appends computed result to its result
      """
      def project(tuples, schema, outcols):
         """
            tuples:  list of location tuples
            schema:  column numbers for each elt in tuple
            outcols: columns to include in output

            returns projection of tuples on outcols
         """
         res = []
         indices = [ i for i in range(len(schema)) if schema[i] in outcols]
         for ltuple in tuples:
            res.append(tuple([ltuple[i] for i in indices]))
         return res 
      def sort_columns(schema, tuples):
         """
            sort columns in tuples by column number
         """
         newres = []
         sschema = sorted(schema)
         for ltuple in tuples:
            t = []
            for col in sschema:
               index = [i for i in range(len(schema)) if schema[i] == col][0]
               t.append(ltuple[index])
            newres.append(tuple(t))
         return sschema, newres      
      def depth_first(ptr,ecount,prev_res,parent):
         found = False
         if len(ptr.children) == 0:
            # a leaf
            if not ptr.done and ptr.ecount == ecount: 
               found = True
               """
               schema:
                  'pred': leaf predicate parameters
                  'tuples': schema (tags) of previous results
                  'tags': 'pred'-'tuples'
               """
               schema = {'pred': ptr.params} 
               new_res = []
               if prev_res != None:
                  # schemas of prev results and predicate params are disjoint
                  disjoint = True
                  partial_newres = []
                  # computation limited with previous results
                  for cols,tuples in prev_res:
                     schema['tuples'] = cols
                     # new columns to add with this predicate
                     schema['tags'] = list(set(ptr.params).difference(set(cols)))
                     if set(schema['tags']) == set(schema['pred']):
                        # leaf predicate's parameters and prev result's tags are disjoint
                        tags = [ self.qtags[i] for i in ptr.params ]
                        partial_newres = partial_newres + self.tagger.select(ptr.op, tags)
                        newschema, newres = cartesian_prod(tuples, partial_newres, schema)
                     else:
                        disjoint = False
                        tags = [ self.qtags[i] for i in schema['tags'] ]
                        newschema, newres = self.tagger._select(ptr.op, tuples, tags, schema)
                     new_res.append((newschema, rm_dups(newres)))
                  if disjoint:
                     ptr.result = [(ptr.params, rm_dups(partial_newres))]
                  else:
                     # project on params and save result in leaf node
                     res = []
                     for cols, tuples in new_res:
                        res = res + project(tuples, cols, ptr.params)
                     ptr.result = [(ptr.params, rm_dups(res))]
               else:
                  # no previous results
                  tags = [self.qtags[col] for col in ptr.params]
                  # get results with columns not sorted
                  newres_unsorted = self.tagger.select(ptr.op, tags)
                  # sort columns
                  newsch, newres = sort_columns(ptr.params, newres_unsorted)
                  new_res = [ (newsch, newres) ]
                  ptr.result = [ (newsch.copy(), newres.copy()) ]
               self.delete_ecount(ecount)
               ptr.done = True
               ptr.ecount = len(ptr.result[0][1])
               # move up new_res to parent
               if parent != None:
                  if parent.op == 'and':
                     # overwrite
                     parent.result = new_res
                  else:
                     # parent is 'or', append result
                     merge_results(parent.result, new_res)
         else:
            # inner node
            if ptr.op == 'and' and len(ptr.result) > 0:
               # and's current results move down
               prev_res = ptr.result
            for child in ptr.children:
               found = depth_first(child, ecount, prev_res, ptr) 
               if found: break
         return found
      min_ecount = self.ecounts[0]
      depth_first(self.root, min_ecount, None, None)
   def execute(self):       ## dab 2022-11-07
      """Execute query
      
      Returns the list of tuple locations that satisfy the query. The number 
      of elements in each tuple is determined from parameters tags and project
      in the constructor. Each tuple has n elements where n is the number
      of elements in project, if project is not [], or the number of elements in tags,
      if project is [].
      """
      self.tokenize()      # partition query into tokens
      self.parse()         # create parse tree
      self.flatten_tree()  # reduce height of tree
      self.update_tree()   # estimate counts of tuples to be evaluated by leaf nodes
      if self.fd != None: self.print_tree()
      while not self.root.done:
         self.compute_leaf() # evaluate leaf with lowest number of tuples to evaluate
         self.update_tree()  # update tree with results of leaf
         if self.fd != None: self.print_tree()
      if len(self.project) == 0:
          oschema = list(range(len(self.qtags)))
      else:
          oschema = self.project
      # project results to oschema
      for i in range(len(self.root.result)):
          self.root.result[i] = project(self.root.result[i], oschema)
      # complete result with cartesian products, if needed
      result = []
      all_true = lambda x: True
      for sch, tuples in self.root.result:
        tags = [self.qtags[oschema[i]] for i in range(len(oschema)) if oschema[i] not in sch]
        if len(tags) == 0:
            result = result + tuples
            continue
        psch = [oschema[i] for i in range(len(oschema)) if oschema[i] not in sch]
        schema = {'tuples': sch, 'pred': psch}
        ptuples = self.tagger.select(all_true, tags)       
        result = result + cartesian_prod(tuples, ptuples, schema)[1]
      # remove duplicates
      sresult = []
      included = set([])
      for r in result:
         if r in included: continue
         sresult.append(r)
         included.add(r)
      return sresult






"""Module that implements functions on location lists

merge_list(Loc list, boolean) -> Loc list
shortest(Loc list) -> Loc list
longest(Loc list) -> Loc list
minus(Loc list, Loc list) -> Loc list
rm_intervals(Loc list, Loc list) -> Loc list
"""
from .loctuple import intersects, subinterval, meets, overlaps, during, starts
from .loctuple import finishes, equal
from .loc import merge, Loc
from .extracterror import handle_error

def merge_list(llist, merge_contiguous=False):
   """Returns the result of merging intersecting/contiguous locations in llist
   
   Parameters:
      llist (Loc list) -- list of locations to be merged
      merge_contiguous (boolean) -- whether contiguous locations would be merged
                                    in addition to intersecting location (default False)
    
   It create a new list by merging intersecting locations in llist. If merge_contiguous 
   is True, it also merges contiguous intervals. Intersecting/contiguous locations must 
   have same offset to merge   
   """
   if len(llist) == 0: return llist 
   slist = sorted(llist, key=lambda x: x.order())
   tomerge = slist[0]
   # merge overlapping intervals
   rem = []
   for intrval in slist[1:]:
       if intersects([tomerge, intrval]) and tomerge.offset == intrval.offset:
          tomerge = merge(tomerge, intrval)
       elif merge_contiguous and meets([tomerge, intrval]):
          tomerge = merge(tomerge, intrval)
       else:
          rem.append(tomerge)
          tomerge = intrval
   rem.append(tomerge)
   return rem
def shortest(llist):
    """Returns the subset of llist with the shortest locations
    
    Parameters:
       llist (Loc list): list of locations to get result from
       
    It returns the subset of locations in llist resulting from removing locations 
    that are supersets of other locations in llist
    """
    if len(llist) == 0: return llist
    slist = sorted(llist, key=lambda x: x.order())
    nlist = [slist[0]]
    for elt in slist[1:]:
        if elt.start() != nlist[len(nlist)-1].start():
            nlist.append(elt)
    i = 0
    res = []
    while i < len(nlist):
        j = i+1
        include = True
        while j < len(nlist):
            if subinterval([nlist[j], nlist[i]]):
                include = False
                break
            j += 1
        if include:
            res.append(nlist[i])
        i += 1
    return res
def longest(llist):
   """Returns the subset of llist with the longest locations
   
   Parameters:
      llist (Loc list): list of locations to get result from
      
   It returns the subset of locations in llist resulting from removing locations 
   that are subsets of other locations in llist
   """
   if len(llist) == 0: return llist
   slist = sorted(llist, key=lambda x: x.order(), reverse=True)
   nlist = [slist[0]]
   for elt in slist[1:]:
      if elt.start() != nlist[-1].start():
         nlist.append(elt)
   i = 0
   res = []
   while i < len(nlist):
      j = i+1
      include = True
      while j < len(nlist):
         if subinterval([nlist[i], nlist[j]]):
            include = False
            break
         j += 1
      if include:
         res.append(nlist[i])
      i += 1
   return sorted(res, key=lambda x: x.order())
def binary_search(llist, trg):
   """
      search <trg> in <llist>
      <trg> is an object of type Loc and <llist> is a
      list of Loc objects
      The order of Loc objects is based on fr, to, offset
      if found, it returns True and index of object that is equal to
      <trg>, otherwise it returns False and index j such that
      llist[j] < trg < llist[j+1] 
   """
   lo = 0
   hi = len(llist)-1
   while lo <= hi:
      mid = int( (lo+hi)/2 )
      if trg.order() == llist[mid].order():
         return (True, mid)
      elif trg.order() < llist[mid].order():
         hi = mid-1
      else:
         lo = mid+1
   return (False, hi)
def minus(llist1, llist2):
    """Returns locations in llist1 that are not in llist2
    
    Parameters:
       llist1 (Loc list) -- locations where result locations come from
       llist2 (Loc list) -- locations to be discarded from result
    """
    if len(llist1) == 0 or len(llist2) == 0:
       return llist1
    res = []
    slist = sorted(llist2, key=lambda x: x.order())
    for loc in llist1:
        isIn, indx = binary_search(slist, loc) 
        if isIn:
            continue
        res.append(loc)
    return res 
def rm_intervals(llist1, llist2):
   """Remove locations in llist2 from llist1
   
   Parameters:
      llist1 (Loc list)
      llist2 (Loc list)
      
   It removes locations in llist2 from llist1. if lx is in llist1 and ly is in llist2
   and lx intersect with ly, remove the intersecting part from lx
   """
   mlist1 = merge_list(llist1)
   mlist2 = merge_list(llist2)
   j = 0
   result = []
   while len(mlist1) != 0:
      loc1 = mlist1.pop(0) 
      (s1, e1, o1) = loc1.order()
      # locate first elt in mlist2 that intersects with loc1
      found = False
      i = j
      while i < len(mlist2):
         loc2 = mlist2[i]
         if intersects([loc1, loc2]):
            #or meets([loc2,loc1]) or before([loc2,loc1]): 
            j = i
            found = True
            break
         i += 1
      if not found:
         result.append(loc1)
         continue
      # remove from loc1 intersecting intervals from mlist2      
      include = True
      for k in range(j,len(mlist2)):
         loc2 = mlist2[k]
         if not intersects([loc1, loc2]): break
         (s1, e1, o1) = loc1.order()
         (s2, e2, o2) = loc2.order()
         if o1 != o2:
            handle_error(110901, 'offsets in rm_intervals parameters do not match')
         if overlaps([loc2,loc1]) or starts([loc2,loc1]):
            loc1 = Loc(e2,e1,o1) 
         elif during([loc2, loc1]):
            loc1 = Loc(s1,s2,o1)
            mlist1.insert(0,Loc(e2,e1,o1))
         elif overlaps([loc1,loc2]) or finishes([loc2,loc1]):
            loc1 = Loc(s1, s2,o1)
         elif equal([loc1, loc2]) or during([loc1, loc2]) or starts([loc1, loc2]) \
              or finishes([loc1,loc2]):
            # excludes loc1
            include = False
            break 
      if include:
         result.append(loc1)
   return result
def main():
   l1 = [(1,4), (7,18), (18,26), (30,40)]
   l2 = [(2,3), (7,10), (11,12), (16,18), (25,35)]
   mlist1 = []
   for (s,e) in l1:
      mlist1.append(Loc(s,e))
   mlist2 = []
   for (s,e) in l2:
      mlist2.append(Loc(s,e))
   result = rm_intervals(mlist1, mlist2)
   print('-----------')
   for loc in result:
      print(loc.order())
if __name__ == "__main__":
   main()



"""Module for tagged text
   
   lit(string) -> literal
   
   Class Tagger Methods:
      SETUP
         __init__(string, boolean)
      TAGGING
         tagRE(string, string, int, boolean)
         tag_loc(string, Loc)
         tag_list(string. Loc list)
         tag_lists(string, list of Loc lists)
         del_tag(string)
      DISPLAY      
         display_matches()
         display_doc()
         display_tag(string)
         display_tuples(list of Loc tuples)
      INSPECT
         get_locs(string/literal, boolean) -> Loc list
         get_text_loc(Loc) -> string
         get_text_tuple(Loc tuple) -> string tuple
         get_text_list(Loc list) -> string list
         get_text_tag(string) -> string list
         in_tag(Loc, string list) -> string
         not_in(Loc list, int pair) -> Loc list
      TRANSFORM   
         project(string, string) -> Loc list
         replace_tag(string, string) -> string
"""
import regex as re

from .extracterror import handle_error
from .loc import Loc, expand
from .loctuple import subinterval
from .loclist import binary_search, merge_list

def lit( str_to_match):
   """
      Returns a literal object for input parameter
   """
   return {'literal': str_to_match}
class Tagger:
   """Tagged text: text and mappings from tags to locations in text
      
   Constructor parameters:
      text (string) -- text to be tagged
      lower_case (boolean) -- if True, convert text to lower case (default True) 
      
   Locations associated with a tag are always sorted by fr,to,offset
   """
   def __init__(self, text, lower_case=True):
      if lower_case:
         self.text = text.lower()
      else:
         self.text = text
      self.spans   = {} # tag --> [ (from, to), ...]
   def tagRE(self, tag, regexp, group=0, overlapped=False):
      """Tag strings in text matching regexp with tag
         
      Parameters:
         tag (string) -- tag
         regexp (string) -- regular expression
         group (int) -- match group within the regular expression (default 0)
         overlapped (boolean) -- whether regular expression matches overlap (default False)
      
      Raises an exception if tag already exists
      """
      if tag in self.spans:
         msg = "Tag {} already in. Did not overwrite".format(tag)
         handle_error(110101, msg)
      self.spans[tag] = [Loc(i[0], i[1]) \
                         for i in self._findpatt(regexp,group,overlapped)]
   def _findpatt(self, pattern,group=0,overlapped=False):
      """
         Returns start and end positions of strings in self.text that match <pattern>
      """
      res = []
      for m in re.finditer(pattern, self.text, overlapped=overlapped):
         fr,to = m.span(group)
         res.append(m.span(group))
      return res
   def display_matches(self):
      """ Prints tagged strings with their respective tags"""
      for tag in self.spans:
         print("{}".format(tag))
         for span in self.spans[tag]:
            offset = span.offset
            fr = offset + span.start()
            to = offset + span.end()
            print("match: {}---".format(self.text[fr:to]))
   def display_tuples(self, ltuples):
      """Display text associated with list of location tuples"""
      for t in ltuples:
         print(self.get_text_tuple(t))
   def tag_loc(self, tag, loc):
      """Tag loc  with tag
         
      Parameters:
         loc (Loc) -- location 
         tag (string) -- tag
         
      Raises a warning if loc is already tagged with tag
      """
      if tag not in self.spans:
         self.spans[tag] = []
      # binary search on whole interval including offset
      found, indx = binary_search(self.spans[tag],loc)
      if not found:
         self.spans[tag] = self.spans[tag][:indx+1] + [loc] + \
                           self.spans[tag][indx+1:]
      else:
         msg = "Tag {} already has location {}"
         handle_error(210102, msg.format(tag, loc.txt_order()))
   def project(self, tag, ref_tag):
      """Converts tag locations to be relative to the beginning of strings tagged as ref_tag
         
      Parameters:
         tag (string) -- tag of locations to be converted
         ref_tag (string) -- reference tag
         
      Exclude locations in tag that are not part of a location in ref_tag.
      Offset of converted locations is first position of corresponding ref_tag location.
      Returns converted locations sorted by (start, end, offset)
      """
      result = []
      taglocs = self.get_locs(tag)
      reftaglocs = self.get_locs(ref_tag)
      for t in taglocs:
         for r in reftaglocs:
            if subinterval([t, r]):
               offset = r.start()
               start = t.start() - offset
               end = t.end() - offset
               result.append(Loc(start, end, offset))
               break
      return sorted(result, key=lambda x: x.order())
   def between(self, startTag, endTag, distance):
      # more efficient than seq_before
      # --- HAS NOT BEEN TESTED ---
      def locateIn(ilist, start):
         """
            locate <start> in <ilist>
            if multiple intervals in <ilist> have <start> as the beginning
            of the interval, returns the index of the first one
            if no interval in <ilist> have <start> as the beginning of 
            interval, returns index i such that <start> is between
            start of ilist[i] and ilist[i+1]
         """
         lo = 0
         hi = len(ilist)-1
         while lo <= hi:
            mid = int( (lo+hi)/2 )
            if start == ilist[mid].start():
               while mid >=0 and start == ilist[mid].start():
                  mid -= 1
               return mid+1
            elif start < ilist[mid].start():
               hi = mid-1
            else:
               lo = mid+1
         return lo
      lower = self.get_locs(startTag)
      upper = self.get_locs(endTag)
      res = []
      for intrval in lower:
         b = locateIn(upper, intrval.end())
         e = locateIn(upper, intrval.end()+distance)
         for indx in range(b, e+1):
            d = upper[indx].start()-intrval.end()
            if d <= distance and d > 0:
               res.append( Loc(intrval.end(), upper[indx].start()) )
      return res
   def tag_list(self, tag, locs):
      """Tag locs with tag
         
      Parameters:
         tag (string) -- tag
         locs (list of Loc) -- locations to be tagged
      """
      for l in locs:
         self.tag_loc(tag, l)
   def tag_lists(self, prefix_tag, loc_lists):
      """Tags lists in loc_lists with prefix_tag_0,..., prefix_tag_n 
      where n=len(loc_lists)-1
          
      Parameters:
         prefix_tag (string) -- prefix of tags to be used
         loc_lists (list of Loc lists) -- lists to be tagged
      """
      for i,llist in enumerate(loc_lists):
         self.tag_list("{}_{}".format(prefix_tag,i),llist)
   def _select(self, pred, tuples, tags, schema):
      """
          select locations tuples from <tuples> and <tags> that satisfy 
          <pred>
          
          pred:   a boolean function on a tuple of locations
          tuples: list of locations tuples
          tags:   list of tags
          schema: a dictionary specifying the column numbers of pred 
                  arguments and tuples 
             'pred':   [i1, ..., ir]
             'tuples': [j1, ..., jt]
             'tags':   [k1, ..., kg]
          output: a list of locations tuples and its schema (col numbers):
             [(loc1, ..., locn), ...] where n=t+g 
             columns in <output> are sorted by column number
          
          creates cartesian product of tuples and locations in tags
          apply <pred> to each element of the cartesian product 
          and output the element if <pred> is True
                  
      """
      def get_pred_args(args, stuples, stags, spred):
         """
            This function extract locations from args to pass as arguments
            of predicate
            
            <args> has the locations to get pred args from:
            args =  [ (l1,...,lt), loc1, ..., locg ]
            <stuples> has column numbers of first element in <args>
            stuples = [j1,...,jt]
            <stags> has column numbers of the other elements in <args>
            stags =               [ k1, ...,  kg]
            <spred> has column numbers of predicate arguments
            spred = [cn1,...,cnr]
         """
         result = []
         if len(stuples) > 0:
            loc_tuple = args[0]
         for col in spred:
            indices = [i for i in range(len(stuples)) if stuples[i] == col ] 
            if len(indices) != 0:
               indx_tuple = indices[0]
               result.append(loc_tuple[indx_tuple])
               continue
            indices = [i for i in range(len(stags)) if stags[i] == col ] 
            indx_tag = indices[0]
            result.append(args[indx_tag+1])
         return tuple(result)
        
      def get_out_tuple(args, stuples, stags, soutput):
         result = []
         if len(stuples) > 0:
            loc_tuple = args[0]
         for col in soutput:
            indices = [i for i in range(len(stuples)) if stuples[i] == col ] 
            if len(indices) != 0:
               indx_tuple = indices[0]
               result.append(loc_tuple[indx_tuple])
               continue
            indices = [i for i in range(len(stags)) if stags[i] == col ] 
            indx_tag = indices[0]
            result.append(args[indx_tag+1])
         return tuple(result)
       
      def check_schema(schema):
         # -------verify constraints-----------
         stuples = []; stags = []; spred = []
         if 'tuples' in schema:
            stuples = schema['tuples']
         if 'pred' in schema:
            spred = schema['pred']
         if 'tags' in schema:
            stags = schema['tags']
         # stuples, stags, and spred must be sets
         if len(stuples) != len(set(stuples)) or len(spred) != len(set(spred)) \
            or len(stags) != len(set(stags)):
            fmt = "Column numbers in schemas must not have duplicates: {},{},{}"
            handle_error(110801, fmt.format(stuples,stags,spred))
         stags_derived = set(spred).difference(set(stuples))
         if stags_derived != set(stags):
            fmt = "Invalid column numbers for tags: {}"
            handle_error(110802, fmt.format(stags))
         # predicate columns come from tuples and tags
         soutput = sorted(stuples + stags)
         return (stuples, stags, spred, soutput)       
      stuples, stags, spred, soutput = check_schema(schema)
      result = []
      nrargs = len(tags)+1      
      locs = [tuples] # list of locations lists
      maxindx = [len(tuples)-1]
      for tag in tags:
          col = self.get_locs(tag)
          if len(col) == 0:
              return (soutput, result) 
          locs.append(col)
          maxindx.append(len(col) - 1)
      indx = [0]*nrargs # current indices for each array ints
      while indx[0] <= maxindx[0]:
          # get tuple of intervals from arrays in ints
          # one from each array
          args = [ locs[i][indx[i]] for i in range(nrargs) ]
          pargs = get_pred_args(args, stuples, stags, spred)
          if pred(pargs):
             otuple = get_out_tuple(args, stuples, stags, soutput)
             result.append(otuple)
          # get next indices
          j = nrargs-1
          while j > 0 and indx[j] == maxindx[j]:
              j -= 1 
          indx[j] += 1
          for k in range(j+1, nrargs):
              indx[k] = 0
      return (soutput, result) 
   def select(self, relation, tags, aggfn=lambda x:x):
      """
         select intervals tagged by <tags> list that satisfy relation
         apply <aggfn> to selected intervals
         output is not sorted
         
         superseded by Query module
      """
      result = []
      nrargs = len(tags)
      maxindx = []
      locs = [] # list of locations lists
      for tag in tags:
         col = self.get_locs(tag)
         if len(col) == 0:
             return result
         locs.append(col)
         maxindx.append(len(col) - 1)
      indx = [0]*nrargs # current indices for each array ints
      while indx[0] <= maxindx[0]:
         # get tuple of intervals from arrays in ints
         # one from each array
         args = tuple([locs[i][indx[i]] for i in range(nrargs)])
         if relation(args):
            result.append(aggfn(args))
         # get next indices
         j = nrargs-1
         while j > 0 and indx[j] == maxindx[j]:
            j -= 1 
         indx[j] += 1
         for k in range(j+1, nrargs):
            indx[k] = 0
      return result

   def not_in(self, tags, refint=None):
      """Get text locations not tagged by tags.
         
      Parameters:
         tags (string list) -- list of tags
         refint (int pair) -- start and end location of text to consider (default None)
                             
      Returns list of locations in text not tagged by any of the tags in 
      parameter tags. If refint is not None, returned locations are
      limited to text[refint[0]:refint[1]]. Returned list is sorted by
      from,to, offset and all offsets are zero.
      """
      if type(tags) is not list:
         handle_error(110104, 'parameter of <not_in> must be a list of tags' )
      if refint == None:
         refint=(0,len(self.text))
      # find intervals to discard
      discard = []
      for tag in tags:
         discard = discard + self.get_locs(tag)
      if len(discard) == 0: return [Loc(0, len(self.text))]
      rem = [expand(loc) for loc in merge_list(discard)]
      
      # build list of locations in <refint> that do not overlap <rem>
      res = []
      point = refint[0]
      for elt in rem:
         loc = expand(elt)
         if loc.start() < point: continue
         if loc.start() >= refint[1]: break
         if point != loc.start():
             res.append(Loc(point, loc.start()))
         point = loc.end()
      if point != refint[1]:
         res.append(Loc(point, refint[1]))
      return sorted(res, key=lambda x: x.order())
   def apply_tags(self, tags):
      """
         show text resulting from replacing text associated with tags 
         in list <tags> by the corresponding tags enclosed by char ~
      """
      txt = self.text
      for tag in self.spans:
         if tag not in tags: continue
         for intrval in self.spans[tag]:
            fr = intrval.start() + intrval.offset
            to = intrval.end() + intrval.offset
            origLen = to-fr
            fill = origLen - len(tag)
            if fill < 0:
               tags = tag[:origLen]
            else:
               tags = tag
            beforeF = int(fill/2)
            afterF = fill - beforeF
            txt = txt[:fr] + '~'*beforeF+ tags + '~'*afterF +txt[to:]
      return txt
   def get_locs(self, tag, overlapped=False):
      """Returns Loc list of text tagged with tag
         
      Parameters:
         tag (string/literal) -- tag or literal
         overlapped (boolean) -- whether string matches overlap when tag is a literal 
                                 (default False)
      """
      if isinstance(tag, dict) and 'literal' in tag:
         # literal
         stag = tag['literal']
         stag = stag.replace('(', '\(').replace(')','\)').replace('.', '\.')
         stag = stag.replace('[', '\[').replace(']','\]').replace('$', '\$')
         stag = stag.replace('^', '\^').replace('?','\?')
         res = [Loc(fr,to) for (fr,to) in self._findpatt(stag,0,overlapped)] 
      elif isinstance(tag, str):
         if tag not in self.spans:
            res = []
         else:
            res = self.spans[tag]
      else:
         handle_error(110105, 'first parameter of get_locs must be a tag or a literal' )        
      return res
   def in_tag(self, loc, tag_list):
      """Returns first tag in tag_list that includes loc, if any, otherwise returns None
         
      Parameters:
         loc (Loc) -- location
         tag_list (string list) -- list of tags
      """
      for tag in tag_list:
         if tag not in self.spans: continue
         # binary search on whole interval
         isIn, indx = binary_search( self.spans[tag], loc )
         if isIn: return tag
      return None
   def replace_tag(self, tag, replacement):
      """Returns text resulting from replacing strings tagged with tag with replacement
         
      Parameters:
         tag (string) -- tag of strings to be replaced
         replacement (string) -- replacement string
      """
      result = self.text
      locs = merge_list(self.get_locs(tag))
      for indx in range(len(locs)-1, -1, -1):
         fr, to = locs[indx].intrval
         offset = locs[indx].offset
         fr = fr + offset
         to = to + offset
         result = result[:fr] + replacement + result[to:]
      return result

   def display_doc(self):
      """display entire text"""
      print(self.text)
   def display_tag(self, tag):
      """Display strings tagged with tag
      
      Parameters:
         tag (string/literal) -- tag name or literal
      """
      ints = self.get_locs(tag)
      for intrval in ints:
         absint = expand(intrval)
         s = absint.start()
         e = absint.end()
         print(self.text[s:e])
   def get_text_loc(self, loc):
      """Returns string in location loc
      
      Parameters:
         loc (Loc) -- location
      """
      absloc = expand(loc)
      s = absloc.start()
      e = absloc.end()
      return self.text[s:e]
   def get_text_list(self, llist):
      """Returns list of strings in locations llist
         
      Parameters:
         llist (Loc list) -- list of locations
      """
      res = []
      absllist = [expand(loc) for loc in llist]
      slist = sorted(absllist, key=lambda x:x.order())
      for loc in slist:
         res.append(self.get_text_loc(loc)) 
      return res
   def get_text_tuple(self, ltuple):
      """Returns tuple of strings in locations ltuple
         
      Parameters:
         ltuple (Loc tuple) -- tuple of locations
      """
      res = []
      for loc in ltuple:
         res.append(self.get_text_loc(loc)) 
      return tuple(res)
   def get_text_tag(self, tag):
      """Returns list of strings tagged by tag
         
      Parameters:
         tag (string) -- tag
      """
      return self.get_text_list(self.get_locs(tag))
   def del_tag(self, tag):
      """Delete tag
      
      Parameters:
         tag (string) -- tag to be deleted
      """
      if tag not in self.spans:
         handle_error(210103, 'Tag {} to remove does not exist'.format(tag))
      else:
         del self.spans[tag]

    



"""Module with common regular expressions and useful functions:

Regular Expressions:
   LINE
   INT
   DECIMAL
   FIELD
   altFIELD
   PAGE
Functions:   
   get_column(string, string) -> Loc list
   cluster(Loc list) -> List of Loc lists
   get_list_span(Loc list) -> Loc
"""

##sys.path.append('/Users/mescobar/home/Professional/Business/py/tagger/src')

from .extracterror import handle_error
from .loc import Loc, expand
from .loctuple import subinterval
from .loctuple import intersects
from .tagger import Tagger, lit

#-------------------- regular expressions --------------------
NEWPAGE = chr(12).encode('ascii')     # new page character as a byte
NEWLINE = b'\n'
PDEL = '($|^|{})'.format(chr(12))     # page delimiter
PCONT = '[^{}]*'.format(chr(12))      # page content
PAGE = '('+PCONT+')' + PDEL
LCONT = '[^\n^{}]*'.format(chr(12))   # line content
LDEL = '(\n|$|^|{})'.format(chr(12))  # line delimiter
LINE = '('+LCONT+')' +LDEL            # line enclosed by line separators
WDEL = '(\n|^|$|{}| )'.format(chr(12))# word delimiter
WORD = '[^\n^{}^ ]+'.format(chr(12))  # word
FIELD = WDEL + '((' + WORD + ' )*' + WORD + ')' # field
DECIMAL = '[0-9,]+(\.[0-9]*){0,1}'  # decimal number with commas
INT = '[0-9,]+'
altFIELD = WDEL + '(' + WORD + ')'            # alternative field
#-------------------- literals ---------------------

def get_column(in_txt, string):
   """Get column in in_text that contains string
   
   Parameters:
      in_txt (string) --  text to get column from
      string (string) --  string in column
   
   Returns text locations of column in in_txt that contains <string>    
   """
   tagged = Tagger(in_txt)
   tagged.tagRE('#FIELD', FIELD, 2, True)
   tagged.tagRE('#LINE', LINE, 1)
   locs = tagged.project('#FIELD', '#LINE')
   tagged.tag_list('#pFIELD', locs)
   tagged.tag_list('#REF', tagged.get_locs(lit(string)))
   cols = cluster(tagged.get_locs('#pFIELD'))
   abs_cols = [list(map(expand, column)) for column in cols] 
   tagged.tag_lists('#COL', abs_cols)
   ncols = len(abs_cols)
   res = []
   for c in range(ncols):
      pair_list = tagged.select(subinterval, ['#REF', '#COL_{}'.format(c)])
      if len(pair_list) > 0:
         res.append(c)
   if len(res) != 1:
      msg = "There is more than one column with text {}".format(string)
      handle_error(110501, msg)
   return (abs_cols[res[0]])

def cluster(llist):
   """Clusters locations in llist into intersecting locations
   
   Parameters:
      llist (Loc list) -- locations to cluster
   
   It creates a list of clusters from llists, where each cluster consists of 
   locations that intersect with at least one other location in that cluster.
   ex: llist = [o1,(1,2), o2,(1,3), o3,(4,5), o4,(4,10)]
       result = [ [o1,(1,2), o2,(1,3)],  [o3,(4,5), o4,(4,10)]]
   Locations in clusters are sorted by (start, end)
   This function ignores the offsets of intervals
   It returns the list of clusters
   """
   clusters = []
   if len(llist) == 0: return clusters
   curr = [llist[0]]
   lo = llist[0].start()
   hi = llist[0].end()
   for t in llist[1:]:
       if intersects([Loc(lo,hi), t]):
           lo = min(lo, t.start())
           hi = max(hi, t.end())
           curr.append(t)
       else:
           clusters.append(curr)
           curr = [t]
           lo = t.start()
           hi = t.end()
   clusters.append(curr)
   return clusters

def categorize(clusters, categories ):
   """
   Parameters
   ----------
   clusters :  a list of clusters (locations lists)
               each cluster is sorted by txt_order()
               clusters = [ cluster_1, .... , cluster_n ] where
               cluster_i = [loc_1, ...., loc_o]
   categories : a list of categories (locations lists)
                each category is sorted by txt_order()
                categories = [ category_1, ... , category_m ]
                category_j = [loc_1, ..., loc_l]

   Returns
   -------
   a list of categorized clusters, one for each category:
      result = [ categorized_clusters_1, ...., categorized_clusters_m]
   each categorized clusters is a list of clusters:
      categorized_clusters_i = [ ccluster_1, ..., ccluster_p]
      ccluster_j = [loc_1, ...., loc_q]
      each loc_k in ccluster_j is a location in cluster_j that is also
      in category_i
      each ccluster_j is sorted by txt_order()

   """
   result = []
   for category in categories:
      categorized_clusters = []
      for cluster in clusters:
         cat = 0
         clu = 0
         ccluster = []
         while cat < len(category) and clu < len(cluster):
            if cluster[clu].txt_order() == category[cat].txt_order():
               ccluster.append(cluster[clu])
               clu += 1
               cat += 1
            elif cluster[clu].txt_order() < category[cat].txt_order():
               clu += 1
            else:
               cat += 1
         categorized_clusters.append(ccluster)
      result.append(categorized_clusters)
   return result
 

def get_list_span(locs):
   """Gets span of all locations in loc
   
   Parameters:
      locs (Loc list) -- locations to get the span from
      
   It returns the span of the intervals in locs, offsets ignored
   """
   if len(locs) == 0: return None
   start = locs[0].start()
   end = locs[0].end()
   for loc in locs[1:]:
      start = min(start, loc.start())
      end = max(end, loc.end())
   return Loc(start, end)

